/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */

/*
* Copyright 2018-2022 Kathy Su, Kay Diederichs
* 
* This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>. 
*/

/**
* Ported from https://doi.org/10.1107/S2059798317000699 by
* @AUTHOR MorellThomas
*/

package jalview.math;

import jalview.bin.Console;

import java.lang.Math;
import java.lang.System;
import java.util.Arrays;
import java.util.ArrayList;

/**
 * A class to model rectangular matrices of double values and operations on them
 */
public class ccAnalysis 
{

  /** TODO
  * DOCUMENT ME
  *
  * @param hSigns ~ <++>
  * @param scores ~ input score matrix
  *
  * @return distrustScores
  */
  private static int[] initialiseDistrusts(byte[] hSigns, MatrixI scores)
  {
    int[] distrustScores = new int[scores.width()];
    
    // loop over symmetric matrix
    for (int i = 0; i < scores.width(); i++)
    {
      byte hASign = hSigns[i];
      int conHypNum = 0;
      int proHypNum = 0;

      for (int j = 0; j < scores.width(); j++)
      {
	double cell = scores.getRow(i)[j];	// value at [i][j] in scores
	byte hBSign = hSigns[j];
	if (!Double.isNaN(cell))
	{
	  byte cellSign = (byte) Math.signum(cell);
	  if (cellSign == hASign * hBSign)
	  {
	    proHypNum++;
	  } else {
	    conHypNum++;
	  }
	}
      }
      distrustScores[i] = conHypNum - proHypNum;
    }
    return distrustScores;
  }

  /** TODO
  * DOCUMENT ME
  *
  * @param hSigns ~ <++>
  * @param distrustScores ~ <++>
  * @param scores ~ input score matrix
  *
  * @return hSigns
  */
  private static byte[] optimiseHypothesis(byte[] hSigns, int[] distrustScores, MatrixI scores)
  {
    // get maximum distrust score
    int[] maxes = MiscMath.findMax(distrustScores);
    int maxDistrustIndex = maxes[0];
    int maxDistrust = maxes[1];

    // if hypothesis is not optimal yet
    if (maxDistrust > 0)
    {
      //toggle sign for hI with maximum distrust
      hSigns[maxDistrustIndex] *= -1;
      // update distrust at same position
      distrustScores[maxDistrustIndex] *= -1;

      // also update distrust scores for all hI that were not changed
      byte hASign = hSigns[maxDistrustIndex];
      for (int NOTmaxDistrustIndex = 0; NOTmaxDistrustIndex < distrustScores.length; NOTmaxDistrustIndex++)
      {
	if (NOTmaxDistrustIndex != maxDistrustIndex)
	{
	  byte hBSign = hSigns[NOTmaxDistrustIndex];
	  double cell = scores.getRow(maxDistrustIndex)[NOTmaxDistrustIndex];

	  // distrust only changed if not NaN
	  if (!Double.isNaN(cell))
	  {
	    byte cellSign = (byte) Math.signum(cell);
	    // if sign of cell matches hypothesis decrease distrust by 2 because 1 more value supporting and 1 less contradicting
	    // else increase by 2
	    if (cellSign == hASign * hBSign)
	    {
	      distrustScores[NOTmaxDistrustIndex] -= 2;
	    } else {
	      distrustScores[NOTmaxDistrustIndex] += 2;
	    }
	  }
	}
      }
      //further optimisation necessary
      return optimiseHypothesis(hSigns, distrustScores, scores);

    } else {
      return hSigns;
    }
  }

  /** TODO
  * DOCUMENT ME
  *
  * runs analysis
  *
  * @param scores ~ score matrix
  *
  * @return
  */
  public static MatrixI run (MatrixI scores)
  {
    MatrixI eigenMatrix = scores.copy();
    try
    {
    /*
    * Calculate correction factor for 2nd and higher eigenvalue(s).
    * This correction is NOT needed for the 1st eigenvalue, because the
    * unknown (=NaN) values of the matrix are approximated by presuming
    * 1-dimensional vectors as the basis of the matrix interpretation as dot
    * products.
    */
    MatrixI scoresOld = scores.copy();
    //&! debug
    System.out.println("input:");
    eigenMatrix.print(System.out, "%8.2f");
    int matrixWidth = eigenMatrix.width(); // square matrix, so width == height
    int matrixElementsTotal = matrixWidth * 2;	//total number of elemts

    int correlationFactor = (matrixElementsTotal - eigenMatrix.countNaN()) / matrixElementsTotal;
    
    /*
    * Calculate hypothetical value (1-dimensional vector) h_i for each
    * dataset by interpreting the given correlation coefficients as scalar
    * products.
    */

    /*
    * Memory for current hypothesis concerning sign of each h_i.
    * List of signs for all h_i in the encoding:
      * *  1: positive
      * *  0: zero
      * * -1: negative
    * Initial hypothesis: all signs are positive.
    */
    byte[] hSigns = new byte[matrixWidth];
    Arrays.fill(hSigns, (byte) 1);

    //Estimate signs for each h_i by refining hypothesis on signs.
    hSigns = optimiseHypothesis(hSigns, initialiseDistrusts(hSigns, eigenMatrix), eigenMatrix);

    //Estimate absolute values for each h_i by determining sqrt of mean of
    //non-NaN absolute values for every row.
    //<++> hAbs = //np.sqrt(np.nanmean(np.absolute(eigenMatrix), axis=1)) 
    double[] hAbs = MiscMath.sqrt(eigenMatrix.absolute().meanRow()); //np.sqrt(np.nanmean(np.absolute(eigenMatrix), axis=1))

    //Combine estimated signs with absolute values in obtain total value for
    //each h_i.
    double[] hValues = MiscMath.elementwiseMultiply(hSigns, hAbs);
    //<++>hValues.reshape((1,matrixWidth));	// doesnt it already look like this

    /*Complement symmetric matrix by using the scalar products of estimated
    *values of h_i to replace NaN-cells.
    *Matrix positions that have estimated values
    *(only for diagonal and upper off-diagonal values, due to the symmetry
    *the positions of the lower-diagonal values can be inferred).
    List of tuples (row_idx, column_idx).*/

    ArrayList<int[]> estimatedPositions = new ArrayList<int[]>();

    // for off-diagonal cells
    for (int rowIndex = 0; rowIndex < matrixWidth - 1; rowIndex++)
    {
      for (int columnIndex = rowIndex + 1; columnIndex < matrixWidth; columnIndex++)
      {
	double cell = eigenMatrix.getValue(rowIndex, columnIndex);
	if (Double.isNaN(cell))
	{
	  //calculate scalar product as new cell value
	  cell = hValues[rowIndex] * hValues[columnIndex];	// something is wrong with hAbs andhValues and everything!!!!!!!!!!!!
          //fill in new value in cell and symmetric partner
	  eigenMatrix.setValue(rowIndex, columnIndex, cell);
	  eigenMatrix.setValue(columnIndex, rowIndex, cell);
	  //save positions of estimated values
	  estimatedPositions.add(new int[]{rowIndex, columnIndex});
	}
      }
    }

    // for diagonal cells
    for (int diagonalIndex = 0; diagonalIndex < matrixWidth; diagonalIndex++)
      {
        double cell = Math.pow(hValues[diagonalIndex], 2);
	eigenMatrix.setValue(diagonalIndex, diagonalIndex, cell);
	estimatedPositions.add(new int[]{diagonalIndex, diagonalIndex});
      }

    /*Refine total values of each h_i:
    *Initialise h_values of the hypothetical non-existant previous iteration
    *with the correct format but with impossible values.
     Needed for exit condition of otherwise endless loop.*/

    double[] hValuesOld = new double[matrixWidth];

    int iterationCount = 0;

    // repeat unitl values of h do not significantly change anymore
    while (true)
    {
      for (int hIndex = 0; hIndex < matrixWidth; hIndex++)
      {
	//@python newH = np.sum(hValues * eigenMatrix[hIndex]) / np.sum(hValues ** 2)
	double newH = Arrays.stream(MiscMath.elementwiseMultiply(hValues, eigenMatrix.getRow(hIndex))).sum() / Arrays.stream(MiscMath.elementwiseMultiply(hValues, hValues)).sum();
	hValues[hIndex] = newH;

      }
      //update values of estimated positions
      for (int[] pair : estimatedPositions)	// pair ~ row, col
      {
        double newVal = hValues[pair[0]] * hValues[pair[1]];
	eigenMatrix.setValue(pair[0], pair[1], newVal);
	eigenMatrix.setValue(pair[1], pair[0], newVal);
      }

      iterationCount++;

      //exit loop as soon as new values are similar to the last iteration
      if (MiscMath.allClose(hValues, hValuesOld, 0d, 1e-05d, false))
      {
        break;
      }

      //save hValues for comparison in the next iteration
      System.arraycopy(hValues, 0, hValuesOld, 0, hValues.length);
    }

    //-----------------------------
    //Use complemented symmetric matrix to calculate final representative
    //vectors.
    //&! debug
    System.out.println("after estimating:");
    eigenMatrix.print(System.out, "%8.2f");

    //Eigendecomposition.
    //eigenMatrix.tqli();

    //&! debug
    //eigenMatrix.print(System.out, "%8.2f");


    } catch (Exception q)
    {
      Console.error("Error computing cc_analysis:  " + q.getMessage());
      q.printStackTrace();
    }
    return eigenMatrix;
  }
}
